require 'fog'
require 'guid'
require 'timeout'
require 'puppet/cloudpack/progressbar'

module Puppet::CloudPack
  module VMware
    class NotFoundError < RangeError; end
    class << self
      def add_timeout_option(action)
        action.option '--timeout=' do
          summary 'Seconds to wait for API calls to complete'
          description <<-EOT
            The maximum number of seconds to wait for API calls to complete.
            Many of the API calls may take considerable time to complete.  If
            not specified, this defaults to #{Fog.timeout} seconds.
          EOT
        end
      end

      def create_connection(options = {})
        Puppet.notice "Connecting ..."
        connection = Fog::Compute[:vsphere]
        Puppet.notice "Connected to #{connection.vsphere_server} as #{connection.vsphere_username} (API version #{connection.vsphere_rev})"
        connection
      end

      def list(options = {})
        timeout = options[:timeout] ? options[:timeout].to_i : Fog.timeout
        connection = create_connection
        if options[:folder] then
          do_in_progress_bar(
            :timeout => timeout,
            :message => "Locating",
            :notice  => "Finding Virtual Machines in Folder #{options[:folder]}"
          ) do
            servers = connection.servers('folder' => options[:folder])
            servers.collect { |s| s.attributes }
          end
        else
          do_in_progress_bar(
            :timeout => timeout,
            :message => "Locating",
            :notice  => "Finding all Virtual Machines ..."
          ) do
            servers = connection.servers
            # Return an Array of Hash objects
            servers.collect { |s| s.attributes }
          end
        end
      end

      def create(options = {})
        timeout = options[:timeout] ? options[:timeout].to_i : Fog.timeout
        if not options[:template] =~ /^\/Datacenters/
          raise ArgumentError, "The --template option must be a path, for example /Datacenters/Solutions/vm/templates/puppet_template."
        end
        total_steps  = 2
        total_steps += 1 if options[:wait_for_boot]
        current_step = 1
        connection = create_connection

        # We expect only one result from find_vm().
        template = find_vm(
          :path       => options[:template],
          :connection => connection,
          :message    => "Locating (#{current_step}/#{total_steps})"
        ).first
        current_step += 1

        new_vm = do_in_progress_bar(
          :timeout => timeout,
          :message => "Starting (#{current_step}/#{total_steps})",
          :notice  => "Starting the clone process"
        ) do
          template.clone('name' => options[:name], 'wait' => options[:wait])
        end
        current_step += 1

        if options[:wait_for_boot] then
          do_in_progress_bar(
            :timeout => timeout,
            :message => "Booting (#{current_step}/#{total_steps})",
            :notice  => "Waiting for the machine to boot and obtain an IP address ..."
          ) do
            new_vm.wait_for(timeout,15) { ipaddress }
          end
        end
        current_step += 1

        { 'status' => 'success' }.merge(filter_server_attributes(new_vm.attributes))
      end

      def stop(options = {})
        validate_path options[:path]
        server = find_vm(:path => options[:path]).first
        do_in_progress_bar(
          :message => "Stopping",
          :notice  => "Stopping Virtual Machine #{server.path}/#{server.name}"
        ) do
          server.stop(:force => options[:force])
        end
      end

      def start(options = {})
        validate_path options[:path]
        server = find_vm(:path => options[:path]).first
        do_in_progress_bar(
          :message => "Starting",
          :notice => "Starting Virtual Machine #{server.path}/#{server.name}"
        ) do
          server.start
        end
      end

      def terminate(options = {})
        validate_path options[:path]
        server = find_vm(:path => options[:path]).first
        if server.power_state =~ /power.*?on/i then
          do_in_progress_bar(
            :message => "Stopping",
            :notice  => "The VM is powered on.  Turning it off."
          ) do
            server.stop(:force => true)
          end
        end
        do_in_progress_bar(
          :message => "Destroying",
          :notice  => "Destroying Virtual Machine #{server.path}/#{server.name}"
        ) do
          server.destroy
        end
      end

      def find(options = {})
        validate_path options[:path]
        found = Array.new
        found << find_vm(:path => options[:path]).map { |vm| vm.attributes }
        found.flatten
      end

      private

      # This method validates a path to a VM.
      # It's import to validate the names before you use take any action on a VM.
      # The find_vm method can return multiple VMs if it's passed just a VM name instead
      # of a path. That's why this method exists.
      def validate_path(path)
        # FIXME: This likely needs great improvement.
        # I need to have more examples of real world setups to do a better regex.
        path =~ /^\// or raise "Invalid path. Must be full path of VM"
      end

      # Filter server attributes which have symbol keys to have string keys
      def filter_server_attributes(attributes)
        attributes.inject({}) { |memo, (k,v)| memo[k.to_s] = v; memo }
      end

      # Find a VM by path.  This is a private method because I suspect there's a much more efficient
      # way to do this and this will be a nice place to refactor.
      # The options hash will initially only support options['path'] but it may make sense to support
      # search filters like EC2 does if we can figure out efficient API calls to VMware.
      def find_vm(options = {})
        # It's really starting to seem like this should be an instance of a class
        # and the connection should be an instance variable.
        connection = options[:connection] || create_connection

        do_in_progress_bar(
          :message => options[:message] || "Locating",
          :notice => "Locating VM at #{options[:path]}"
        ) do
          paths   = options[:path].split('/')
          vm_folder = paths[0...-1].join('/')
          vm_name = paths[-1]

          found = if vm_folder.empty?
            connection.servers.find_all { |s|  s.name == vm_name }
          else
            # This is a semi-optimized strategy to find a VM in a subset of the
            # inventory.
            connection.servers('folder' => vm_folder).find_all { |s|  s.name == vm_name and s.path == vm_folder }
          end

          if found.empty?
            raise NotFoundError, "Could not find VM at #{options[:path]}"
          end

          #Ensure we return an array
          found.to_a
        end
      end

      # Take a block and a timeout and display a progress bar while we're doing our thing
      def do_in_progress_bar(options = {}, &blk)
        timeout = options[:timeout] ? options[:timeout].to_i : Fog.timeout
        start_time = Time.now
        abort_time = start_time + timeout

        Puppet.notice "#{options[:notice]} (Started at #{start_time.strftime("%I:%M:%S %p")})"
        eta_msg = if (timeout <= 120) then
                    "#{timeout} seconds at #{abort_time.strftime("%I:%M:%S %p")}"
                  else
                    "#{timeout / 60} minutes at #{abort_time.strftime("%I:%M %p")}"
                  end
        Puppet.notice "Control will be returned to you in #{eta_msg} if #{options[:message].downcase} is unfinished."

        progress_bar = Puppet::CloudPack::ProgressBar.new(options[:message], timeout)
        progress_mutex = Mutex.new

        progress_thread = Thread.new do
          loop do
            progress = Time.now - start_time
            progress_mutex.synchronize { progress_bar.set progress }
            sleep 0.5
          end
        end

        block_return_value = nil
        begin
          Timeout.timeout(timeout) do
            block_return_value = blk.call
          end
        ensure
          progress_mutex.synchronize { progress_bar.finish; progress_thread.kill }
        end
        end_time = Time.now
        block_return_value
      end
    end
  end
end
