define(['dom', 'sandbox', './request', '../collections/requests', 'text!../templates/requests.html', 'core_lib/alerts'],
  function($, sandbox, RequestView, RequestsCollection, template, Alerts) {

  var EmptyView = sandbox.mvc.ItemView({
    tagName: 'tr',
    template: '<td colspan="99">There are no certificates to accept or reject.</td>'
  });

  var RequestsView = sandbox.mvc.CompositeView({

    itemView: RequestView,
    emptyView: EmptyView,

    itemViewContainer: 'tbody',

    template: template,

    events: {
      'click #certificates-clear-requests': 'clear_requests',
      'click #certificates-accept-all': 'accept_all',
      'click #certificates-reject-all': 'reject_all'
    },

    ui: {
      controls: '#certificates-controls',
      open_count: '.certificates-open-csr-count',
      clear_requests: '#certificates-clear-requests',
      accept_all: '#certificates-accept-all',
      reject_all: '#certificates-reject-all',
      list_message: '#certificates-list-message'
    },

    initialize: function(options) {

      this.collection = options.collection || new RequestsCollection;

      this.bindTo(this.collection, 'change:status', this.update_ui_controls, this);
      this.bindTo(this.collection, 'remove', this.update_clear_completed, this);

      var self = this;

      $(document).on('click', function(event) {
        var target = $(event.target),
            in_tooltip = target.parents('.popover, .tooltip');

        // if we've clicked inside a tooltip do nothing, allow double clicks/link clicks to work as normal
        if (in_tooltip.length > 0) {
          return true;
        }

        // if the click was on an element contained within a tooltip toggler replace the click target
        // with the container
        var in_link = target.parents('a[rel="tooltip"], a[rel="popover"]');
        if (in_link.length > 0) {
          target = in_link;
        }

        if (target.is('a[rel="tooltip"]')) {
          // if the click was on a tooltip, hide all other tooltips and popovers
          self.$('a[rel="tooltip"]').not(target).tooltip('hide');
          self.$('a[rel="popover"]').popover('hide');
          return true;
        }

        if (target.is('a[rel="popover"]')) {
          // if the click was on a popover, hide all other popovers and tooltips
          self.$('a[rel="popover"]').not(target).popover('hide');
          self.$('a[rel="tooltip"]').tooltip('hide');
          return true;
        }

        // otherwise hide all tooltips and popovers
        self.hide_tooltips();
      })
    },

    onRender: function() {
      this.update_ui_controls();
    },

    /**
     * Keep the total request count and accept/reject all button count in sync.
     */
    update_ui_controls: function() {
      this.update_open_count();
      this.update_clear_completed();
      this.update_action_buttons();
    },

    /**
     * Remove all completed requests from the UI
     */
    clear_requests: function() {
      this.collection.clear_requests();
    },

    reject_all: function() {
      this._do_multi_action('reject_all');
    },

    accept_all: function() {
      this._do_multi_action('accept_all');
    },

    _do_multi_action: function(action) {
      this.hide_error();
      this.disable_controls();

      var reenable = $.proxy(this.enable_controls, this),
          update = $.proxy(this.update_ui_controls, this),
          show_error = $.proxy(this.show_error, this);

      this.collection[action]().done([reenable, update])
                               .fail([reenable, show_error]);
    },

    disable_controls: function() {
      this.ui.controls.find('button').attr('disabled', 'disabled');
    },

    enable_controls: function() {
      this.ui.controls.find('button').removeAttr('disabled');
    },

    show_error: function(message) {
      var error = new Alerts.error({
        content: message,
        classes: ['certificates-list-error']
      });

      this.ui.list_message.html(error.render().el).show();
    },

    hide_error: function() {
      this.ui.list_message.html('').hide();
    },

    /**
     * Hides all open tooltips. Do this before any major UI rearrangement as the tooltips are
     * not anchored to their parents.
     */
    hide_tooltips: function() {
      this.$('a[rel="tooltip"]').tooltip('hide');
      this.$('a[rel="popover"]').popover('hide');
    },

    update_open_count: function() {
      var pending = this.collection.open_requests().length;

      this.ui.open_count.text('Pending node requests: ' + pending);

      $('a.node-requests-link').text('node requests (' + pending + ')');
    },

    update_clear_completed: function() {
      var complete = this.collection.completed_requests().length,
          text;

      if (complete === 1) {
        text = 'Clear 1 accepted/rejected request';
        this.ui.clear_requests.removeAttr('disabled');
      } else if (complete > 1) {
        text = 'Clear ' + complete + ' accepted/rejected requests';
        this.ui.clear_requests.removeAttr('disabled');
      } else {
        text = 'There are no accepted/rejected requests to clear';
        this.ui.clear_requests.attr('disabled', 'disabled');
      }

      this.ui.clear_requests.text(text);
    },

    update_action_buttons: function() {
      var actionable = this.collection.actionable_requests().length;

      if (actionable === 0) {
        this.ui.accept_all.attr('disabled', 'disabled');
        this.ui.reject_all.attr('disabled', 'disabled');
      } else {
        this.ui.accept_all.removeAttr('disabled');
        this.ui.reject_all.removeAttr('disabled');
      }
    }
  });

  return RequestsView;
});